{% load datatable_tags %}
{% load ajax_helpers %}


<div class="card report-card">
    <div class="card-header">
        <div class="d-flex align-items-center">
            <h5 class="mr-auto">{{ title }}</h5>
            {% block button_menu %}{{ menus.button_menu.render }}{% endblock %}
        </div>
    </div>
    <div class="dashboard-tile detail p-0">
        <div class="content" style="padding: 2px;">
            <div class="col-xs-12" style="overflow: auto;">
                <div class="alert alert-warning m-0" role="alert">
                    {{ error_message|safe }}
                </div>
            </div>
        </div>
    </div>
</div>
