# Generated by Django 5.1.3 on 2025-11-21 14:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0019_multicellstyle_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='multicellstyle',
            name='font_size',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
    ]
