# Generated by Django 5.1.3 on 2025-11-25 22:46

import django.db.models.deletion
import time_stamped_model.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0020_alter_multicellstyle_font_size'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultiValueHeldQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('query', models.JSONField(blank=True, null=True)),
                ('multi_value_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.multivaluereport')),
                ('report_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
            ],
            options={
                'get_latest_by': 'created',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='multivaluereportcell',
            name='multi_value_held_query',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='advanced_report_builder.multivalueheldquery'),
        ),
    ]
