# Generated by Django 5.1.3 on 2025-07-04 14:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0017_alter_barchartreport_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportquery',
            name='target',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='advanced_report_builder.target'),
        ),
        migrations.AddField(
            model_name='target',
            name='period_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Daily'), (2, 'Weekly'), (3, 'Monthly'), (4, 'Quarterly')], default=3),
        ),
    ]
